package de.hawhamburg.reachability.reasoner.engine;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.apache.log4j.Logger;

import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.conf.ReasonerConfig;
import de.hawhamburg.reachability.conf.enumeration.ConfigType;
import de.hawhamburg.reachability.enumeration.ReachabilityGroup;
import de.hawhamburg.reachability.enumeration.ReachabilityStatus;
import de.hawhamburg.reachability.json.JsonReachabilitySensorData;
import de.hawhamburg.reachability.reasoner.Reasoner;
import de.hawhamburg.reachability.reasoner.ReasonerAgent;
import de.hawhamburg.reachability.util.OutputUtil;
import de.hawhamburg.reachability.util.PropertiesUtil;


/** 
 * An example reasoning engine to explain the basic reasoner implementation.<br><br>
 * 
 * This {@link Reasoner} implementation is an example reasoner implementation for explanation reasons. It is able to 
 * handle active user feedback from user interface. Internal sensor data will be supplied at runtime through the 
 * {@link #setSensorData(JsonReachabilitySensorData)} method. The reasoner estimates the user reachability and provides
 * it by {@link #getEstimatedReachability()}.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ExampleReasoner extends Reasoner {

    /**
     * The name of the property which contains the time for how long active feedback shall be minded.
     */
    private static final String MIND_FEEDBACK_KEY   = "REASONING_ENGINE_MIND_FEEDBACK";

    /**
     * Basic (user-defined) properties.
     */
    private Properties reasonerProps                = null;
    
    /**
     * Starting time of the reasoner.
     */
    private long startingAt                         = 0;

    /**
     * Create a new example reasoning engine.
     */
    public ExampleReasoner() {
        String propSource   = ReasonerConfig.get("REASONING_ENGINE_PROPERTY_SOURCE");
        this.startingAt     = System.currentTimeMillis();
        this.reasonerProps  = PropertiesUtil.getProperties(propSource);
    }

    /**
     * Adds a new sensor object or replaces the last known of this dimension. The dimension and the sensor object can be
     * extracted from the {@link JsonReachabilitySensorData} object. This method will be called by the super class 
     * {@link Reasoner} at runtime if a new internal sensor data arrives.
     * 
     * @param data      The {@link JsonReachabilitySensorData} which has been measured.
     */
    protected final void setSensorData(final JsonReachabilitySensorData data) {
        // Write info output
        String vals = null;
        for (Map.Entry<Integer, Double> x : data.Values.entrySet()) {
            if (vals == null) {
                vals = x.getKey() + "x" + x.getValue();
            } else {
                vals += ", " + x.getKey() + "x" + x.getValue();
            }
        }
        OutputUtil.info("Receiving new SensorData: " + data.Dimension + " values: " + vals);
        
        // ### use the data to calculate the reachability
    }

    /**
     * Supplies the actual estimated {@link ReachabilityStatus} to all {@link ReachabilityGroup}s as {@link Map} of
     * {@link ReachabilityStatus} to {@link ReachabilityGroup}. This method will be called by the super class
     * {@link Reasoner} at runtime to publish the actual estimated reachability.
     * 
     * @return          A {@link Map} of estimated {@link ReachabilityStatus} .
     */
    public final Map<ReachabilityGroup, ReachabilityStatus> getEstimatedReachability() {
        // Initialize a new map
        Map<ReachabilityGroup, ReachabilityStatus> result = new HashMap<ReachabilityGroup, ReachabilityStatus>();
        
        // Fetch the estimated reachability state of each reachability group
        for (ReachabilityGroup p : ReachabilityGroup.values()) {
            
            // Use the local getEstimatedReachability(ReachabilityGroup) method to determine the actual estimated state
            result.put(p, getEstimatedReachability(p));
        }
        
        // Return the result
        return result;
    }

    /**
     * Supplies the actual estimated {@link ReachabilityStatus} to a specific {@link ReachabilityGroup}. 
     * 
     * @param group     The {@link ReachabilityGroup} for which the estimated {@link ReachabilityStatus} shall be 
     *                  returned.
     * @return          The {@link ReachabilityStatus} related to the requested {@link ReachabilityGroup}.
     */
    public final ReachabilityStatus getEstimatedReachability(final ReachabilityGroup group) {
        // ### Calculate the estimated reachability state
        
        return ReachabilityStatus.NOT_CALCULATED;
    }

    /**
     * Handle active user feedback which sets a certain {@link ReachabilityGroup} to a specified 
     * {@link ReachabilityStatus}. This method will be called by the super class {@link Reasoner} at runtime whenever
     * new active user feedback arrives.
     * 
     * @param group     The {@link ReachabilityGroup} for which the feedback has been given.
     * @param status    The user expected {@link ReachabilityStatus} to this {@link ReachabilityGroup}.
     */
    public void handleFeedback(final ReachabilityGroup group,
                               final ReachabilityStatus status) {
        // ### handle active user feedback
    }

    /**
     * Returns reasoner defined options for a specific property. This may be useful for dynamic property options. This
     * method will be called by the property service of the super class {@link Reasoner} if the type for the property 
     * was defined as <code>AGENT_DEFINED</code>
     * 
     * @param property  The requested property as {@link String}.
     * @return          The values of the requested property as {@link Map} of {@link String} to {@link String}.
     */
    public final Map<String, String> getValuesForProperty(final String property) {
        return new HashMap<String, String>();
    }

    /**
     * Loads and returns the current local properties of a specified {@link ConfigType}. This method will be called by
     * the super class {@link Reasoner} at runtime if the properties of this reasoner are requested by a user interface.
     * 
     * @param type      Loads the current configuration of a certain type defined as {@link ConfigType}.
     * @return          The requested configuration as {@link Properties} object.
     */
    public final Properties loadCurrentProperties(final ConfigType type) {
        Properties props = new Properties();
        
        switch(type) {
            case REASONER:
              String propSource = ReasonerConfig.get("REASONING_ENGINE_PROPERTY_SOURCE");
              props             = PropertiesUtil.getProperties(propSource);
            default:
                break;
        }
        
        return props;
    }

    /**
     * Update local properties. This method will be called by the super class {@link Reasoner} at runtime if a property 
     * change from the user interface or actual configuration data of other components arrive.
     * 
     * @param newProps  The actual defined and correct {@link Properties} set.
     * @param changed   The changed {@link Properties} set.
     * @param type      The properties type which shall be updated as {@link ConfigType}.
     */
    public final void updateProperties(final Properties newProps,
                                       final Properties changed,
                                       final ConfigType type) {
        switch(type) {
            case REASONER:
                
                // ### Handle changes...
                
                // Save new properties to file
                String propSource = ReasonerConfig.get(PROPERTY_FILE_SOURCE_KEY);
                PropertiesUtil.save(this.reasonerProps, propSource);
                
                // Notify other systems
                this.pushProperties();
                break;
            case SYSTEM:
                MainConfig.updateProperties(newProps);
                break;
            default:
                break;
        }
        
    }

    /**
     * Write current state to log and give all resources free. This method will be called by the super class 
     * {@link Reasoner} at runtime if this service will be ended.
     */
    protected final void cleanUp() {
        this.dump();
        
        // ### Free resource...
    }

    /**
     * Write current state and configuration of this reasoning engine to log...
     */
    public void dump() {
        // ### Write reasoner state and current configuration to log...
    }
    
    /**
     * Main procedure to start and run this reasoning engine as {@link ReasonerAgent}.
     * 
     * @param args      The system supplied start arguments as array of {@link String}.
     */
    public static void main(final String[] args) {
        // Remove all existing loggers
        Logger.getRootLogger().removeAllAppenders();
        
        // Start this reasoning engine as reasoner agent.
        new ReasonerAgent(new ExampleReasoner());
    }
}
